#include		<stdio.h>
#include		<stdlib.h>
#include		<string.h>
#include		<ctype.h>
#include		<Assert.h>

#include		"redir.h"

/****************************************************************************
 *                                                                          *
 *       C-Phantom Network INT 2F Function 11 Redirector - cphantom.c       *
 *                                                                          *
 *     Copyright (c) 1992-1995 Printing Communications Associates, Inc.     *
 *                                                                          *
 *                            Thomas F. Divine                              *
 *                          4201 Brunswick Court                            *
 *                       Smyrna, Georgia 30080 USA                          *
 *                           TEL: 770 436-1714                              *
 *                           FAX: 770 436-7246                              *
 *                           CIS: 71077,1525                                *
 *                           AppleLink: PCA                                 *
 *                                                                          *
 ****************************************************************************/

#define VERSION "1.00B5"

/*
 * $Workfile:   cphantom.c  $
 *  $Logfile:   D:/USR/SRC/NET/AFP/CPHANTOM/VCS/CPHANTOM.C_V  $
 *   $Author:   PCA-TFD-SCSI  $
 *     $Date:   11 Jun 1992 11:56:56  $
 * $Revision:   1.7  $
 *
 * $Log:   D:/USR/SRC/NET/AFP/CPHANTOM/VCS/CPHANTOM.C_V  $
 * 
 *    Rev 1.7   11 Jun 1992 11:56:56   PCA-TFD-SCSI
 * 1.) Modification to assembler glue to allow operation under Windows 3.1.
 * 2.) Added support for SeekFromEnd.
 * 
 *    Rev 1.6   15 May 1992 13:32:10   PCA-TFD-SCSI
 * 1.) New version V1.00B4.
 * 2.) Added INT 61h control stub.
 * 3.) Modifications to support compilation in COMPACT memory model.
 * 
 *    Rev 1.5   13 May 1992 23:53:44   PCA-TFD-SCSI
 * 1.) New version V1.00B3.
 * 2.) Added timestamp support.
 * 3.) Added support for the following functions:
 *     CloseFile        - Subfunction 06h
 *     Read             - Subfunction 08h
 *     Write            - Subfunction 09h
 *     OpenFile         - Subfunction 16h
 *     CreateFile       - Subfunction 17h
 *     SpecialOpen      - Subfunction 2Eh
 * 
 *    Rev 1.4   11 May 1992 10:42:48   PCA-TFD-SCSI
 * 1.) New version 1.00B2.
 * 2.) Added routines VCATFindFirst() and VCATFindNext().
 * 3.) Simplified doXXXX routines by using VCATFindFirst() and VCATFindNext().
 * 
 *    Rev 1.3   09 May 1992 11:13:10   PCA-TFD-SCSI
 * 1.) New version V1.00B1.
 * 2.) Extensive revision. Generic redirector functions moved to 2fsubs.c.
 * 3.) Added support for MkDir, SetFileAttrib, GetFileAttrib, and RenameFile.
 * 4.) Improved dummy volume catalog.
 * 
 *    Rev 1.2   06 May 1992 18:16:24   PCA-TFD-SCSI
 * 1.) New version 1.00A3.
 * 2.) Added support for ChDir and DeleteFile.
 * 3.) Added support for different DOS versions. Tested on DOS 3.30 and
 *     DOS 5.00.
 * 
 *    Rev 1.1   06 May 1992 09:21:30   PCA-TFD-SCSI
 * 1.) New version 1.00A2.
 * 2.) Added support for RmDir.
 * 3.) Modifications to MakeTemplate, MatchTemplate, FindFirst and FindNext.
 * 
 *    Rev 1.0   05 May 1992 18:24:10   PCA-TFD-SCSI
 * Initial revision.
 */

/*                                 CPHANTOM
 *
 * This program is based upon information from the following excellent
 * work, and all who contributed thereto:
 *
 *                             UNDOCUMENTED DOS:
 * A Programmer's Guide to Reserved MS-DOS Functions and Data Structures
 *             by Andrew Schulman, Raymond J. Michels, Jim Kyle,
 *                 Tim Paterson, David Maxey, and Ralf Brown
 *(Reading MA: Addison-Wesley, 1990, 694 pp., $39.95, ISBN 0-201-57064-5)
 */

/* Undocumented DOS includes Pascal source for an example DOS file system
 * redirector using INT 2F, function 11h. Unfortunately, Pascal is not
 * a usable language for TFD.
 *
 * This program is a partial translation of the PHANTOM.PAS redirector
 * example from the Pascal to the C programming language. It is not
 * complete. Basically, it represents the point where the author decided
 * to quit experimenting and get on with the real thing. Only the following
 * functions work:
 *
 *		Installation Check - Subfunction 00h
 *		RmDir              - Subfunction 01h
 *		MkDir              - Subfunction 03h
 *		ChDir              - Subfunction 05h
 *		CloseFile          - Subfunction 06h
 *		Read               - Subfunction 08h
 *		Write              - Subfunction 09h
 *		GetDiskSpace       - Subfunction 0Ch
 *		SetFileAttrib      - Subfunction 0Eh
 *		GetFileAttrib      - Subfunction 0Fh
 *		RenameFile         - Subfunction 11h
 *		DeleteFile         - Subfunction 13h
 *    OpenFile           - Subfunction 16h
 *    CreateFile         - Subfunction 17h
 *		FindFirst          - Subfunction 1Bh
 *		FindNext           - Subfunction 1Ch
 *    SeekFromEnd        - Subfunction 21h
 *		SpecialOpen        - Subfunction 2Eh
 *
 *
 * CPHANTOM is placed in the public domain in the hope that others with
 * more experiance than TFD can offer improvements and add support for
 * functions which are stubbed out herein. Hopefully, these contributions
 * will be of use to TFD and other developers who are attempting to develop
 * new system-level products.
 * 
 * See the enclosed README.TXT file for additional information.
 *
 * For a while, anyway, I'll incorporate changes and re-post them on CIS.
 *
 * Happy redirecting!
 *
 * Thomas F. Divine
 */


/****************************************************************************
 *                          Canned Volume Catalog                           *
 ****************************************************************************/

/* Canned Director Drive Number And Letter
------------------------------------------ */
static char		drive_letter;	// Our Drive Letter
static char		drive_no;		// Our Drive Number (A: is 1, etc.)


/* Canned Volume Catalog Entires
-------------------------------- */
#define	MAX_ENTRIES		20

/*
 * In the canned volume catalog example, index into the VCAT array plus
 * VCAT_BASE is the File/Directory ID(FlDrID). Each file and directory
 * therefore has a unique ID number in the range from VCAT_BASE to
 * MAX_ENTRIES + VCAT_BASE.
 *
 * The FlDrID of 2 is defined as the ROOT.
 *
 * The start_clstr field in the VCAT array is used to contain the
 * parent FlDrID.
 */
#define	ROOT				2					// AFP-Style
#define	VCAT_BASE		3

typedef
struct NodeParms
{
	Uint8		fname[11];
	Uint8		fattr;							// DOS Attributes
	Uint16	time_lstupd;
	Uint16	date_lstupd;
	Uint16	ParentID;						// This Node's Parent ID
	Uint16	FlDrID;							// This Node's Unique ID
	Uint32	FlLgLen;							// Logical Length Of Data File
}
	NodeParms;

/* The Volume Catalog(VCAT)
--------------------------- */
static NodeParms VCAT[ MAX_ENTRIES + 1 ] =
{
	//=============== Initial Canned Volume Catalog Entries ===============

	// FlDrID == 3; ParentID == 11 (Fruit)
	{ "Apple      ", _A_SUBDIR, 0x0800, 0x18A1, 11, 3, 0L },

	// FlDrID == 4; ParentID == 11 (Fruit)
	{ "Bananna    ", _A_SUBDIR, 0x0820, 0x18A1, 11, 4, 0L },

	// FlDrID == 5; ParentID == 12 (Fruit)
//	{ "Cherry  C  ", _A_NORMAL | _A_ARCH, 0x0840, 0x18A2, 2, 5, 1234L },
	{ "Printer    ", _A_NORMAL | _A_ARCH, 0x0840, 0x18A2, 2, 5, 1234L },

	// FlDrID == 6; ParentID == 7 (Vegs)
	{ "Beans   TXT", _A_NORMAL | _A_HIDDEN, 0x0860, 0x18A3, 7, 6, 4567L },

	// FlDrID == 7; ParentID == 2 (ROOT)
	{ "Vegs       ", _A_SUBDIR, 0x0880, 0x18A1, 2, 7, 0L },

	// FlDrID == 8; ParentID == 7 (Vegs)
	{ "Carrot     ", _A_RDONLY, 0x08A0, 0x18A5, 7, 8, 77777L },

	// FlDrID == 9; ParentID == 2
	{ "Grape   SYS", _A_NORMAL | _A_HIDDEN | _A_SYSTEM,
		0x08C0, 0x18A1, 2, 9, 55555L },

	// FlDrID == 10; ParentID == 2
	{ "Hominy  DOC", _A_NORMAL, 0x08F0, 0x18A1, 2, 10, 99999L },

	// FlDrID == 11; ParentID == 2 (ROOT)
	{ "Fruit      ", _A_SUBDIR, 0x0800, 0x18A5, 2, 11, 0L },

	// FlDrID == 12; ParentID == 2
	{ "Paper      ", _A_SUBDIR, 0x0820, 0x18A4, 2, 12, 0L },

	//==================== Empty Volume Catalog Entries ==================

	// FlDrID == 13
	{ "", 0, 0, 0, 0, 13, 0L }, 

	// FlDrID == 14
	{ "", 0, 0, 0, 0, 14, 0L },

	// FlDrID == 15
	{ "", 0, 0, 0, 0, 15, 0L },

	// FlDrID == 16
	{ "", 0, 0, 0, 0, 16, 0L },

	// FlDrID == 17
	{ "", 0, 0, 0, 0, 17, 0L },

	// FlDrID == 18
	{ "", 0, 0, 0, 0, 18, 0L },

	// FlDrID == 19
	{ "", 0, 0, 0, 0, 19, 0L },

	// FlDrID == 20
	{ "", 0, 0, 0, 0, 20, 0L },

	// FlDrID == 21
	{ "", 0, 0, 0, 0, 21, 0L },

	// FlDrID == 22
	{ "", 0, 0, 0, 0, 22, 0L },

	// FlDrID == 23
	{ "BOZO       ", _A_NORMAL | _A_ARCH, 0, 0x18A1, 0, 23, 121212L }
};


/* Map A ParentID And A FCB Template To The Item's Own Unique FlDrID
-------------------------------------------------------------------- */
NodeParms *VCATFindFirst( Uint16 ParentID, Uint8 *pTemplate )
{
	int				index, match;

	/* Sanity Checks
	---------------- */
	/* Walk The Volume Catalog
	-------------------------- */
	for( index = 0; index < MAX_ENTRIES; ++index )
	{
		/* Don't Examine Empty Catalog Entries
		-------------------------------------- */
		if( *VCAT[ index ].fname == 0 )
		{
			continue;
		}

		if( *pTemplate == 0 && VCAT[ index ].FlDrID == ParentID )
		{
			return( &VCAT[ index ] );
		}

		if( VCAT[ index ].ParentID == ParentID )
		{
			match = fsMatchTemplate( pTemplate, VCAT[ index ].fname );

			if( match )
			{
				return( &VCAT[ index ] );
			}
		}
	}

	return( (NodeParms * )NULL );						// Not Found
}


/* Find The Next Item Matching A ParentID And A FCB Template
------------------------------------------------------------ */
NodeParms *VCATFindNext( Uint16 FlDrID, Uint16 ParentID, Uint8 *pTemplate )
{
	int			index, match;

	/* Sanity Checks
	---------------- */
	assert( pTemplate );

	/* Walk The Volume Catalog, Beginning After FlDrID
	-------------------------------------------------- */
	for( index = FlDrID - VCAT_BASE + 1; index < MAX_ENTRIES; ++index )
	{
		/* Don't Examine Empty Catalog Entries
		-------------------------------------- */
		if( *VCAT[ index ].fname == 0 )
		{
			continue;
		}

		if( *pTemplate == 0 && VCAT[ index ].FlDrID == ParentID )
		{
			return( &VCAT[ index ] );
		}

		if( VCAT[ index ].ParentID == ParentID )
		{
			match = fsMatchTemplate( pTemplate, VCAT[ index ].fname );

			if( match )
			{
				return( &VCAT[ index ] );
			}
		}
	}

	return( (NodeParms * )NULL );						// Not Found
}


/* Distill The Directory Component Of A Split Path To A Unique Dir FlDrID
-------------------------------------------------------------------------
 * Examples of the split directory component of a path are:
 *
 *     "\\"           -> FlDrID 2 (i.e., ROOT)
 *     "\\Vegs\\"     -> FlDrID 7
 */
Uint16 VCATDistillDir( Uint16 ParentID, Uint8 *pSplitDir )
{
	static char		TempBuff[ 128 ];
	static Uint8	Template[ 11 ];
	Uint8				*p;
	NodeParms		*nodePtr;

	/* Sanity Checks
	---------------- */
	assert( pSplitDir );

	strcpy( (char * )TempBuff, (char * )pSplitDir );

	p = (Uint8 * )strtok( (char * )TempBuff, "\\" );

	while( p )
	{
		/* Make A FCB-Style Template From The Directory Component
		--------------------------------------------------------- */
		fsMakeTemplate( Template, p );

		nodePtr = VCATFindFirst( ParentID, Template );

		if( !nodePtr )
		{
			break;
		}

		ParentID = nodePtr->FlDrID;

		p = (Uint8 * )strtok( NULL, "\\" );
	}

	return( ParentID );
}


/* Get A Free NodeParms Entry
----------------------------- */
NodeParms *VCATGetFreeNode( void )
{
	int			index;
	NodeParms	*nodePtr;

	/* Walk The Catalog
	------------------- */
	for( index = 0; index < MAX_ENTRIES; ++index )
	{
		/* Free Entries Have Empty fname Field
		-------------------------------------- */
		if( *VCAT[ index ].fname == 0 )
		{
			nodePtr = &VCAT[ index ];

			nodePtr->FlDrID = index + VCAT_BASE;	// Init The FlDrID Field

			return( &VCAT[ index ] );
		}
	}

	return( (NodeParms * )NULL );
}


/****************************************************************************
 *               Interrupt 2F Function 11h Subfunction Handlers             *
 ****************************************************************************/

/* GENERAL
----------
 * The following operations have been performed before calling any of
 * the INT 2F Function 11h Subfunction handlers:
 *
 * 1.) The fsCallForUs() function has been used to verify that the call
 *     is for our redirected drive.
 *
 * 2.) The global REGSET pointer, rsPtr, has been initialized.
 *
 * The following additional operations have been performed before calling
 * those functions which use SDA_FN1:
 *
 * 1.) SDA_FN1 has been verified to be for our drive.
 * 2.) SDA_FN1 has been split into components.
 * 3.) A FCB-style search template has been built from the filename
 *     component of SDA_FN1.
 *
 * The following additional operations have been performed before calling
 * this functions which access the SFT:
 *
 * 1.) sftPtr has been setup to point to the SFT associated with the file.
 */

/* Installation Check - Subfunction 00h
--------------------------------------- */
static int doInstCheck( void )
{
	rsPtr->x.ax = 0x00FF;					// Installed
	return( 0 );								// Call Processed
}


/* Get Disk Space - Subfunction 0Ch
----------------------------------- */
static int doGetDiskSpace( void )
{
	/* Just a simple canned response. Could make it dynamic - but
	 * why bother in an example?
	 */

	/* Basic Cluster Size - AX * CX = Cluster Size
	---------------------------------------------- */
	rsPtr->x.ax = 1;							// Sectors Per Cluster
	rsPtr->x.cx = 4096;						// Bytes Per Sector


	/* Total Volume Size = AX * CX * BX
	----------------------------------- */
	rsPtr->x.bx = 10000;						// Total Clusters


	/* Volume Free Space = AX * CX * DX
	----------------------------------- */
	rsPtr->x.dx = 250;						// Unused Clusters

	return( 0 );								// Call Processed
}


/* Remove Directory - Subfunction 01h
--------------------------------------- */
static int doRmDir( void )
{
	NodeParms	*nodePtr;
	int			j, found;
	Uint16		ParentID;

	/* Search For Matching Directories
	---------------------------------- */
	found = 0;

	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	while( nodePtr )
	{
		/* Check Directories Only
		------------------------- */
		if( !( nodePtr->fattr & _A_SUBDIR ) )
		{
			break;			// Path Specified A File, Not A Directory
		}
		else
		{
			/*
			 * Should make additional checks here. i.e., is directory
			 * empty, is it the current directory, etc.
			 */

			/* Verify That Directory Is Empty
			--------------------------------- */
			for( j = 0; j < MAX_ENTRIES; ++j )
			{
				if( *VCAT[ j ].fname != 0
					&& VCAT[ j ].ParentID == nodePtr->FlDrID )
				{
					fsFailCall( EACCES );	// Access Denied
					return( 0 );				// Call Processed
				}
			}

			/* Delete The Directory By Zapping The fname
			-------------------------------------------- */
			nodePtr->fname[ 0 ] = '\0';

			found = 1;
		}

		nodePtr = VCATFindNext( nodePtr->FlDrID, ParentID, FN1Template );
	}

	if( !found )
	{
		fsFailCall( ENOPATH );				// Path Not Found
	}

	return( 0 );								// Call Processed
}


/* Make Directory - Subfunction 03h
----------------------------------- */
static int doMkDir( void )
{
	int			index;
	Uint16		ParentID;
	NodeParms	*nodePtr;

	/* Search For Anything With Matching Name
	----------------------------------------- */
	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	if( nodePtr )
	{
		/* Name Already In Use
		---------------------- */
		fsFailCall( EACCES );			// Access Denied
		return( 0 );						// Call Processed
	}

	/* Search For Empty Directory Entry
	----------------------------------- */
	for( index = 0; index < MAX_ENTRIES; ++index )
	{
		if( *VCAT[ index ].fname == 0 )
		{
			/* Initialize The Empty Directory Entry
			--------------------------------------- */
			memcpy( VCAT[ index ].fname, FN1Template, 11 );
			VCAT[ index ].fattr = _A_SUBDIR;

			/* Assign Current Date/Time
			--------------------------- */
			VCAT[ index ].date_lstupd = fsDate();
			VCAT[ index ].time_lstupd = fsTime();

			VCAT[ index ].FlDrID = index + VCAT_BASE;
			VCAT[ index ].ParentID = ParentID;
			VCAT[ index ].FlLgLen = 0;

			return( 0 );						// Call Processed
		}
	}

	fsFailCall( EACCES );					// Access Denied
	return( 0 );								// Call Processed
}


/* Change Directory - Subfunction 05h
------------------------------------- */
static int doChDir( void )
{
	int			found;
	Uint16		ParentID;
	NodeParms	*nodePtr;

	/* Sanity Checks
	---------------- */
	assert( FN1FileName );

	/* Special Handling For The Root
	-------------------------------- */
	if( *FN1FileName == 0 && strcmp( (char * )FN1Dir, "\\" ) == 0 )
	{
		return( 0 );								// CD To Root Is Always OK!
	}

	/* Search For Matching Directories
	---------------------------------- */
	found = 0;

	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	if( nodePtr )
	{
		/* Check Directories Only
		------------------------- */
		if( nodePtr->fattr & _A_SUBDIR )
		{
			found = 1;
		}
	}

	if( !found )
	{
		fsFailCall( ENOPATH );					// Path Not Found
	}

	return( 0 );									// Call Processed
}


/* Close File - Subfunction 06h
------------------------------- */
static int doCloseFile( void )
{
	Uint16		UseCount;
	NodeParms	*nodePtr;

	/* Sanity Checks
	---------------- */
	assert( sftPtr );

	UseCount = fsDecSFTUseCount( sftPtr );

	if( UseCount == 1 )
	{
		sftPtr->handle_cnt = 0;
		sftPtr->dir_sector = 0;
		sftPtr->devdrv_ptr = NULL;
	}

	/* Update Modification Time, If File Is Dirty
	--------------------------------------------- */
	if( !( sftPtr->dev_info & 0x40 ) )
	{
		nodePtr = &VCAT[ sftPtr->dir_entryno - VCAT_BASE ];

		nodePtr->date_lstupd = fsDate();
		nodePtr->time_lstupd = fsTime();
	}

	return( 0 );
}


/* Random Read - Subfunction 08h
--------------------------------
 * On entry, SDA_DTA points to buffer to be filled and rsPtr->x.cx contains
 * the requested read count. Note that rsPtr->x.cx is Uint16, and can have
 * large values (e.g., 0xFE00). This function must fill SDA_DTA with the
 * requested number of bytes unless end-of-file is reached.
 *
 * Note that this is a random read with offset specified by sftPtr->f_pos.
 * For example, DOS implements a REWIND or SEEK is implemented simply
 * setting sftPtr->f_pos to a new value before making the next read.
 */
static int doRead( void )
{
	NodeParms	*nodePtr;

	/* Sanity Checks
	---------------- */
	assert( sftPtr );

	/* Verify That sftPtr->dir_entryno Is Valid
	------------------------------------------- */
	if( sftPtr->dir_entryno < VCAT_BASE
		|| sftPtr->dir_entryno - VCAT_BASE > MAX_ENTRIES )
	{
		rsPtr->x.cx = 0;

		// SHOULD FAIL CALL???
		return( 0 );
	}

	nodePtr = &VCAT[ sftPtr->dir_entryno - VCAT_BASE ];

	/* Read On File
	--------------- */
	if( sftPtr->f_pos + rsPtr->x.cx > nodePtr->FlLgLen )
	{
		if( nodePtr->FlLgLen - sftPtr->f_pos > 0 )
		{
			rsPtr->x.cx = nodePtr->FlLgLen - sftPtr->f_pos;
		}
		else
		{
			rsPtr->x.cx = 0;
		}
	}

	sftPtr->f_pos += rsPtr->x.cx;

	return( 0 );
}


/* Random Write - Subfunction 09h
---------------------------------
 * On entry, SDA_DTA points to buffer to be written and rsPtr->x.cx contains
 * the requested write count. Note that rsPtr->x.cx is Uint16, and can have
 * large values (e.g., 0xFE00). This function must write all bytes at SDA_DTA
 * unless an error is encountered.
 */
static int doWrite( void )
{
	NodeParms	*nodePtr;

	/* Sanity Checks
	---------------- */
	assert( sftPtr );

	/* Verify That sftPtr->dir_entryno Is Valid
	------------------------------------------- */
	if( sftPtr->dir_entryno < VCAT_BASE
		|| sftPtr->dir_entryno - VCAT_BASE > MAX_ENTRIES )
	{
		rsPtr->x.cx = 0;

		// SHOULD FAIL CALL???
		return( 0 );
	}

	nodePtr = &VCAT[ sftPtr->dir_entryno - VCAT_BASE ];

	/* Can't Write To Read-Only Files
	--------------------------------- */
	if( nodePtr->fattr & _A_RDONLY )
	{
		fsFailCall( EACCES );		// Access Denied
		return( 0 );					// Call Processed
	}

	/* Write To File
	---------------- */
	sftPtr->f_pos += rsPtr->x.cx;

	if( sftPtr->f_pos > nodePtr->FlLgLen )
	{
		nodePtr->FlLgLen = sftPtr->f_pos;
	}

	sftPtr->f_size = nodePtr->FlLgLen;

	sftPtr->dev_info &= ~0x40;				// Mark File As Dirty

	nodePtr->fattr |= _A_ARCH;				// Set Archive Bit

//	rsPtr->x.cx = rsPtr->x.cx;				// Report Bytes Actually Written

	return( 0 );
}


/* Set File Attribute - Subfunction 0Eh
--------------------------------------- */
static int doSetFileAttrib( void )
{
	NodeParms	*nodePtr;
	Uint16		ParentID;

	/* Sanity Checks
	---------------- */
	assert( FN1Dir );
	assert( FN1Template );

	/* Search For Matching Files
	---------------------------- */
	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	if( nodePtr )
	{
		/* UNDOCUMENTED DOS Says New Attrib On Stack. I Say In CX!
		---------------------------------------------------------- */
		nodePtr->fattr = rsPtr->x.cx & 0x2F;
		return( 0 );
	}

	fsFailCall( ENOPATH );				// Path Not Found

	return( 0 );								// Call Processed
}


/* Get File Attribute And Size - Subfunction 0Fh
------------------------------------------------ */
static int doGetFileAttrib( void )
{
	int			found;
	Uint16		ParentID;
	NodeParms	*nodePtr;

	/* Sanity Checks
	---------------- */
	assert( FN1Dir );
	assert( FN1Template );

	/* Search For Matching Files
	---------------------------- */
	found = 0;

	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	if( nodePtr )
	{
		/* Return Attributes In AX
		-------------------------- */
		rsPtr->x.ax = nodePtr->fattr;

		/* Return File Size In BX:DI
		---------------------------- */
		rsPtr->x.bx = (Uint16 )( ( nodePtr->FlLgLen >> 16L ) & 0xFFFFL );
		rsPtr->x.di = (Uint16 )( nodePtr->FlLgLen & 0xFFFFL );

		found = 1;
	}

	if( !found )
	{
		fsFailCall( ENOPATH );				// Path Not Found
	}

	return( 0 );								// Call Processed
}


/* Rename File - Subfunction 11h
--------------------------------
 * Note that source and destination names may contain wildcards. This means
 * that the search on the source must match on wildcards. It also means
 * that the destination name is built in a special way; in particular,
 * only destination characters which are not wild are changed. Think of
 * how the DOS command "rename *.c *.cpp" would be implemented by a single
 * call to this function.
 */
static int doRenameFile( void )
{
	int			j, found;
	Uint16		OldParentID, NewParentID;
	Uint8			NewName[11];
	NodeParms	*nodePtr, *dupNodePtr;

	/* Sanity Checks
	---------------- */
	assert( FN1Dir );
	assert( FN1Template );

	/* Search For Matching Files
	---------------------------- */
	found = 0;

	OldParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( OldParentID, FN1Template );

	while( nodePtr )
	{
		/* Check Files Only
		------------------- */
		if( !( nodePtr->fattr & ( _A_VOLID | _A_SUBDIR ) ) )
		{
			/* Can't Rename A Read-Only File
			-------------------------------- */
			if( nodePtr->fattr & _A_RDONLY )
			{
				fsFailCall( EACCES );		// Access Denied
				return( 0 );					// Call Processed
			}

			/* Make Sure The New Name Is Not In Use
			--------------------------------------- */
			fsSplitPath( SDA_FN2, FN2Prefix, FN2Drive, FN2Dir, FN2FileName );


			/* Make A FCB-Style Template From SDA_FN2's FileName Component
			-------------------------------------------------------------- */
			fsMakeTemplate( FN2Template, FN2FileName );

				
			/* Find The New Parent ID
			------------------------- */
			NewParentID = VCATDistillDir( 2, FN2Dir );


			/* Build The New Name
			--------------------- */
			memcpy( NewName, nodePtr->fname, 11 );

			for( j = 0; j < 11; ++j )
			{
				if( FN2Template[ j ] != '?' )
				{
					NewName[ j ] = FN2Template[ j ];
				}
			}

			/* Search The Catalog, Looking For Collisions With New Name
			----------------------------------------------------------- */
			dupNodePtr = VCATFindFirst( NewParentID, NewName );

			if( dupNodePtr )
			{
				/* Name Already In Use
				---------------------- */
				fsFailCall( EACCES );			// Access Denied
				return( 0 );						// Call Processed
			}

			/* Rename The File
			------------------ */
			memcpy( nodePtr->fname, NewName, 11 );

			found = 1;
		}

		nodePtr = VCATFindNext( nodePtr->FlDrID, OldParentID, FN1Template );
	}

	if( !found )
	{
		fsFailCall( ENOPATH );				// Path Not Found
	}

	return( 0 );								// Call Processed
}


/* Delete File - Subfunction 13h
--------------------------------
 * Note the while loop in this function. In some cases, FN1 may contain
 * wildcards; in such cases, this function must delete all files which
 * match the wildcards.
 */
static int doDeleteFile( void )
{
	NodeParms	*nodePtr;
	int			found;
	Uint16		ParentID;

	/* Sanity Checks
	---------------- */
	assert( FN1Dir );
	assert( FN1Template );

	/* Search For Matching Files
	---------------------------- */
	found = 0;

	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	while( nodePtr )
	{
		/* Check Files Only
		------------------- */
		if( !( nodePtr->fattr & ( _A_VOLID | _A_SUBDIR ) ) )
		{
			/* Can't Delete A Read-Only File
			-------------------------------- */
			if( nodePtr->fattr & _A_RDONLY )
			{
				fsFailCall( EACCES );			// Access Denied
				return( 0 );						// Call Processed
			}

			/* Delete The File By Zapping The fname
			--------------------------------------- */
			nodePtr->fname[ 0 ] = '\0';

			found = 1;
		}

		nodePtr = VCATFindNext( nodePtr->FlDrID, ParentID, FN1Template );
	}

	if( !found )
	{
		fsFailCall( ENOPATH );				// Path Not Found
	}

	return( 0 );								// Call Processed
}


/* Open Existing File - Subfunction 16h
--------------------------------------- */
static int doOpenFile( void )
{
	int			found;
	Uint16		ParentID;
	NodeParms	*nodePtr;

	/* Sanity Checks
	---------------- */
	assert( FN1Dir );
	assert( FN1Template );

	/* Search For Matching Files
	---------------------------- */
	found = 0;

	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	while( nodePtr )
	{
		/* Check Files Only
		------------------- */
		if( !( nodePtr->fattr & ( _A_VOLID | _A_SUBDIR ) ) )
		{
			/* Initialize The Supplied SFT Entry
			------------------------------------ */
			memcpy( sftPtr->filename, nodePtr->fname, 11 );
			sftPtr->open_mode = SDA_OPEN_MODE & 0x7F;
			sftPtr->attr_byte = nodePtr->fattr;
			sftPtr->f_size = nodePtr->FlLgLen;
			sftPtr->f_pos = 0L;
			sftPtr->dev_info = 0x8040 | drive_no;
			sftPtr->dir_entryno = nodePtr->FlDrID;
			sftPtr->devdrv_ptr = NULL;
			fsSetSFTOwner( sftPtr );

			found = 1;

			break;
		}

		nodePtr = VCATFindNext( nodePtr->FlDrID, ParentID, FN1Template );
	}

	if( !found )
	{
		fsFailCall( ENOPATH );				// Path Not Found
	}

	return( 0 );								// Call Processed
}


/* Create/Truncate File - Subfunction 17h
-----------------------------------------
 * If a file exists with matching name, then it is truncated.
 */
static int doCreateFile( void )
{
	Uint16		ParentID;
	NodeParms	*nodePtr;

	/* Sanity Checks
	---------------- */
	assert( FN1Dir );
	assert( FN1Template );

	/* Search For Anything With Matching Name
	----------------------------------------- */
	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	if( nodePtr )
	{
		/* If The Matching Node Isn't A File, Then Fail The Create
		---------------------------------------------------------- */
		if( nodePtr->fattr & ( _A_VOLID | _A_SUBDIR ) )
		{
			/* Name Already In Use
			---------------------- */
			fsFailCall( EACCES );			// Access Denied
			return( 0 );						// Call Processed
		}

		/* Can't Truncate A Read-Only File
		-------------------------------- */
		if( nodePtr->fattr & _A_RDONLY )
		{
			fsFailCall( EACCES );			// Access Denied
			return( 0 );						// Call Processed
		}
	}

	/* Find Empty Directory Entry
	----------------------------- */
	if( !nodePtr )
	{
		nodePtr = VCATGetFreeNode();
	}

	if( nodePtr )
	{
		/* Initialize The Empty Directory Entry
		--------------------------------------- */
		nodePtr->ParentID = ParentID;
		memcpy( nodePtr->fname, FN1Template, 11 );
		nodePtr->fattr = 0;

		/* Assign Current Date/Time
		--------------------------- */
		nodePtr->date_lstupd = fsDate();
		nodePtr->time_lstupd = fsTime();

		nodePtr->FlLgLen = 0;

		/* Initialize The Supplied SFT Entry
		------------------------------------ */
		memcpy( sftPtr->filename, nodePtr->fname, 11 );
		sftPtr->open_mode = 0x01;			// Assume An Open Mode
		sftPtr->attr_byte = _A_ARCH;		// Should Be From Stack!!!!!!!!!!!
		sftPtr->f_size = nodePtr->FlLgLen;
		sftPtr->f_pos = 0L;
		sftPtr->dev_info = 0x8040 | drive_no;
		sftPtr->dir_entryno = nodePtr->FlDrID;
		sftPtr->devdrv_ptr = NULL;
		fsSetSFTOwner( sftPtr );

		return( 0 );							// Call Processed
	}

	fsFailCall( EACCES );					// Access Denied
	return( 0 );								// Call Processed
}


/* Find First - Subfunction 1Bh
-------------------------------
 * The simplicity of the CPhantom enumeration scheme may be misleading. It
 * works because of the simplicity of the Volume Catalog implementation.
 *
 * In a more realistic implementation, the developer must provide an
 * implementation which supports:
 *		1.) multiple concurrent enumerations(e.g. TREE or recursive moves),
 *		2.) situations where files may be deleted (or added or renamed)
 *		    between sequential calls to FindNext (e.g., a wildcard deletion).
 *
 * This can be interesting...
 */
static int doFindFirst( void )
{
	Uint16		ParentID;
	NodeParms	*nodePtr;

	/* Handle Volume Label Search
	----------------------------- */
	if( SDA_SRCH_ATTR & _A_VOLID )
	{
		/* Initialize The Search Data Block(SDB)
		---------------------------------------- */
		SDA_SDB.drv_lett = drive_letter | 0x80;
		SDA_SDB.dir_entry = 0;
  		memcpy( SDA_SDB.srch_tmpl, FN1Template, 11 );


		/* Make Sure That The Following Label Fits In 11 Bytes!
		------------------------------------------------------- */
		sprintf( (char * )SDA_FOUND_FILE.fname, "%s%c", SIGNATURE, drive_letter );

		SDA_FOUND_FILE.date_lstupd = fsDate();
		SDA_FOUND_FILE.time_lstupd = fsTime();

		SDA_FOUND_FILE.fattr = _A_VOLID;
		SDA_FOUND_FILE.fsiz = 0;

		return( 0 );									// Call Processed
	}
	else
	{
		/*
		 * For FindFirst, sdaPtr->fn1 is a fully-qualified "truename" path of
		 * the following example forms:
		 *
		 * 	CPhantom-G:\TestDir\Lemon\????????.???
		 * 	CPhantom-G:\TestDir\Lemon\????????
		 * 	CPhantom-G:\TestDir\Lemon\????????.C
		 * 	CPhantom-G:\TestDir\Lemon\SUNKIST.TXT
		 *
		 * The tail may be explicit or may contain wildcards.
		 *
		 * In addition, sdaPtr->srch_attr contains the search attributes
		 * which must be matched.
		 */

		ParentID = VCATDistillDir( 2, FN1Dir );

		nodePtr = VCATFindFirst( ParentID, FN1Template );

		while( nodePtr )
		{
			/* Match Against Search Attributes
			---------------------------------- */
			if( ( nodePtr->fattr & 0x1E ) == 0
				|| nodePtr->fattr & SDA_SRCH_ATTR )
			{
				/* Mark The Search Data Block(SDB) As Updated
				--------------------------------------------- */
				SDA_SDB.drv_lett = drive_letter | 0x80;

				/* Initialize Search Template
				----------------------------- */
				memcpy( SDA_SDB.srch_tmpl, FN1Template, 11 );

				/* Use These Fields To Save Info We Need For FindNext
				----------------------------------------------------- */
				SDA_SDB.par_clstr = nodePtr->ParentID;
				SDA_SDB.dir_entry = nodePtr->FlDrID;


				/* Update found_file Directory Entry Record
				------------------------------------------- */
				strcpy( (char * )SDA_FOUND_FILE.fname,
					(char * )nodePtr->fname );
				strupr( (char * )SDA_FOUND_FILE.fname );

				SDA_FOUND_FILE.time_lstupd = nodePtr->time_lstupd;
				SDA_FOUND_FILE.date_lstupd = nodePtr->date_lstupd;

				SDA_FOUND_FILE.fattr = nodePtr->fattr;
				SDA_FOUND_FILE.fsiz = nodePtr->FlLgLen;

				return( 0 );
  			}

			nodePtr = VCATFindNext( nodePtr->FlDrID, ParentID, FN1Template );
		}
	}

	fsFailCall( ENMFILE );				// No More Files
	return( 0 );							// Call Processed
}


/* Find Next - Subfunction 1Ch
------------------------------ */
static int doFindNext( void )
{
	NodeParms	*nodePtr;

	if( !(SDA_SRCH_ATTR & _A_VOLID) )
	{
		/* Search The Directory
		----------------------- */
		nodePtr = VCATFindNext( SDA_SDB.dir_entry, SDA_SDB.par_clstr,
			SDA_SDB.srch_tmpl );

		while( nodePtr )
		{
			/* Match Against Search Attributes
			---------------------------------- */
			if( ( nodePtr->fattr & 0x1E ) == 0
				|| nodePtr->fattr & SDA_SRCH_ATTR )
			{
				/* Mark The Search Data Block(SDB) As Updated
				--------------------------------------------- */
				SDA_SDB.drv_lett = drive_letter | 0x80;

				/* Use These Fields To Save Info We Need For FindNext
				----------------------------------------------------- */
				SDA_SDB.par_clstr = nodePtr->ParentID;
				SDA_SDB.dir_entry = nodePtr->FlDrID;

				/* Update found_file Directory Entry Record
				------------------------------------------- */
				strcpy( (char * )SDA_FOUND_FILE.fname,
					(char * )nodePtr->fname );
				strupr( (char * )SDA_FOUND_FILE.fname );

				SDA_FOUND_FILE.time_lstupd = nodePtr->time_lstupd;
				SDA_FOUND_FILE.date_lstupd = nodePtr->date_lstupd;

				SDA_FOUND_FILE.fattr = nodePtr->fattr;
				SDA_FOUND_FILE.fsiz = nodePtr->FlLgLen;

				return( 0 );
			}

			nodePtr = VCATFindNext( nodePtr->FlDrID, SDA_SDB.par_clstr,
				SDA_SDB.srch_tmpl );
		}
	}

	fsFailCall( ENMFILE );					// No More Files
	return( 0 );								// Call Processed
}


/* Seek From End Of File - Subfunction 21h
------------------------------------------ */
static int doSeekFromEnd( void )
{
	Uint32	Offset;

	/* CX:DX Contains Offset (in bytes) From End
	-------------------------------------------- */
	Offset = (rsPtr->x.cx << 16) | rsPtr->x.dx;


	/* Move To New Offset
	--------------------- */
	sftPtr->f_pos = sftPtr->f_size + Offset;


	/* Extend File Size
	------------------- */
	if( sftPtr->f_pos > sftPtr->f_size )
	{
		sftPtr->f_size = sftPtr->f_pos;
	}

	return( 0 );
}


/* Multi-Function Open - Subfunction 2Eh
----------------------------------------
 * Called in DOS 4.XX+ only.
 */
static int doSpecialOpen( void )
{
	Uint16		ParentID, Result;
	NodeParms	*nodePtr;

	Result = 0;

	/* Search For Matching Files
	---------------------------- */
	ParentID = VCATDistillDir( 2, FN1Dir );

	nodePtr = VCATFindFirst( ParentID, FN1Template );

	if( nodePtr )
	{
		/* Can Only Open Files!
		----------------------- */
		if( nodePtr->fattr & ( _A_VOLID | _A_SUBDIR ) )
		{
			fsFailCall( EACCES );				// Access Denied
			return( 0 );							// Call Processed
		}

		/* Action If File Exists
		------------------------ */
		if( ( SDA_SPOP_ACT & 0x0F ) == 0 )
		{
			fsFailCall( EACCES );				// Access Denied
			return( 0 );							// Call Processed
		}

		/* For Open Or Truncate, Check Access Mode
		------------------------------------------ */
		if( ( SDA_SPOP_ACT & 0x03 ) != 0 )
		{
			/* Check Access Mode
			-------------------- */
			switch( SDA_SPOP_MODE & 0x03 )
			{
				case 0:								// Read-Only
					break;

				case 1:								// Write-Only
				case 2:								// Read/Write
					if( nodePtr->fattr & _A_RDONLY )
					{
						fsFailCall( EACCES );	// Access Denied
						return( 0 );				// Call Processed
					}
					break;

				default:
					break;
			}
		}

		if( ( SDA_SPOP_ACT & 0x01 ) != 0 )
		{
			/* Open The File
			---------------- */
			Result |= 0x01;						// File Opened
		}
		else if( ( SDA_SPOP_ACT & 0x02 ) != 0 )
		{
			/* Can't Truncate An Open File
			------------------------------ */

			/* Open The File
			---------------- */

			/* Truncate The File
			-------------------- */
			nodePtr->FlLgLen = 0;				// That's All For CPhantom...

			Result |= 0x03;						// File Replaced
		}
		else
		{
			/* Invalid Action - Wasn't Fail, Open, Or Truncate
			-------------------------------------------------- */

			nodePtr = (NodeParms * )NULL;
		}
	}
	else
	{
		/* Action If File Does Not Exist
		-------------------------------- */
		if( ( SDA_SPOP_ACT & 0xF0 ) == 0 )
		{
			fsFailCall( EACCES );	// Access Denied
			return( 0 );				// Call Processed
		}
		else if( ( SDA_SPOP_ACT & 0x10 ) != 0 )
		{
			/* Find Empty Directory Entry
			----------------------------- */
			nodePtr = VCATGetFreeNode();

			if( !nodePtr )
			{
				fsFailCall( 0x82 );				// Ran Out Of Directory Entries
				return( 0 );						// Call Processed
			}

			/* Initialize The Empty Directory Entry
			--------------------------------------- */
			nodePtr->ParentID = ParentID;
			memcpy( nodePtr->fname, FN1Template, 11 );
			nodePtr->fattr = SDA_SPOP_ATTR;

			/* Assign Current Date/Time
			--------------------------- */
			nodePtr->date_lstupd = fsDate();
			nodePtr->time_lstupd = fsTime();

			nodePtr->FlLgLen = 0;

			Result |= 0x02;						// File Created
		}
		else
		{
			/* Invalid Action Wasn't Fail Or Create
			--------------------------------------- */
			nodePtr = (NodeParms * )NULL;
		}
	}

	if( nodePtr )
	{
		/* Initialize The Supplied SFT Entry
		------------------------------------ */
		memcpy( sftPtr->filename, nodePtr->fname, 11 );
		sftPtr->open_mode = SDA_SPOP_MODE & 0x7F;
		sftPtr->attr_byte = nodePtr->fattr;
		sftPtr->f_size = nodePtr->FlLgLen;
		sftPtr->f_pos = 0L;
		sftPtr->dev_info = 0x8040 | drive_no;
		sftPtr->dir_entryno = nodePtr->FlDrID;
		sftPtr->devdrv_ptr = NULL;
		fsSetSFTOwner( sftPtr );

		rsPtr->x.dx = Result;
	}

	if( !nodePtr )
	{
		fsFailCall( ENOPATH );				// Path Not Found
	}

	return( 0 );								// Call Processed
}


/****************************************************************************
 *                       Redirector Control Functions                       *
 ****************************************************************************/

/*
 * In many cases there is a need to have a private interface to perform
 * operations beyond the scope of those defined by INT 2F function 11h.
 *
 * A prime example of such a function is an "unload" function - which
 * is not implemented in CPhantom. CPhantom supports only one private
 * function - a private installation check.
 */

/* Redirector Control Function
------------------------------ */
void doControl( void )
{
	/* Dispatch On Function In AH
	----------------------------- */
	switch( rsPtr->h.ah )
	{
		/* Private Commands - Function C0h
		---------------------------------- */
		case 0xC0:
			/* Dispatch On Subfunction In AL
			-------------------------------- */
			switch( rsPtr->h.al )
			{
				/* Private Installation Check - Subfunction 00h
				----------------------------------------------- */
				case 0:
					rsPtr->h.al = 0xFF;					// Installed
					break;

				default:
					break;
			}
		break;

		default:
			break;
	}
}

/****************************************************************************
 *                   C-Language Interrupt Function Dispatcher               *
 ****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

/* Redirector C-Language Interrupt Entry, Called From ASM Glue
--------------------------------------------------------------
 * Return value of 0 indicated that the command was handled by this
 * redirector. A non-zero return value indicates that the interrupt
 * should be chained to the next handler in the chain.
 *
 * This entry point is called for software interrupts:
 *
 *		2F, Fcn 11h            - For network redirector functions.
 *		6X, Any Fcn EXCEPT 11h - For propritary redirector control functions.
 */
int far Redirector( REGSET rs )
{
	/* Setup Global Register Set Pointer
	------------------------------------ */
	rsPtr = &rs;


	/* Call Redirector Control Function
	----------------------------------- */
	if( rs.h.ah != 0x11 )
	{
		doControl();
		return( 0 );
	}

	/* ES:DI Points To SFT For File
	-------------------------------
	 * This is not valid for calls which do not use the SFT, but it is
	 * easy to do it once here.
	 */
	sftPtr = (SFT31 * )MK_FP( rsPtr->x.es, rsPtr->x.di );


	/* Verify That This Call Is For Us
	---------------------------------- */
	if( !fsCallForUs( drive_no ) )
	{
		return( 1 );								// Pretend We Didn't See It
	}

	/* Additional Work For Calls Which Use SDA_FN1
	---------------------------------------------- */
	switch( rs.h.al )
	{
		/*
		 * 1.) Verify that SDA_FN1 is for our drive.
		 * 2.) Split SDA_FN1 into components.
		 * 3.) Make a search template from the filename component
		 *     of SDA_FN1.
		 */
		case rdrRmDir:
		case rdrMkDir:
		case rdrChDir:
		case rdrSetFileAttrib:
		case rdrGetFileAttrib:
		case rdrRenameFile:
		case rdrDeleteFile:
		case rdrOpenFile:
		case rdrCreateFile:
		case rdrFindFirst:
		case ifsSpecialOpen:
			if( fsOurDrive( SDA_FN1 ) != drive_no )
			{
				fsFailCall( ENOPATH );				// Path Not Found
				return( 0 );							// Call Processed
			}

			/* Split SDA_FN1 Into Components
			-------------------------------- */
			fsSplitPath( SDA_FN1, FN1Prefix, FN1Drive, FN1Dir, FN1FileName );

			/* Make A FCB-Style Template From SDA_FN1's FileName Component
			-------------------------------------------------------------- */
			fsMakeTemplate( FN1Template, FN1FileName );

			break;

		default:
			break;
	}


	/* Dispatch On Subfunction In AL
	-------------------------------- */
	switch( rs.h.al )
	{
		/* Installation Check - Subfunction 00h
		--------------------------------------- */
		case rdrInstCheck:
			return( doInstCheck( ) );


		/* Remove Directory - Subfunction 01h
		--------------------------------------- */
		case rdrRmDir:
			return( doRmDir( ) );


		/* Make Directory - Subfunction 03h
		----------------------------------- */
		case rdrMkDir:
			return( doMkDir( ) );


		/* Change Directory - Subfunction 05h
		------------------------------------- */
		case rdrChDir:
			return( doChDir( ) );


		/* Close File - Subfunction 06h
		------------------------------- */
		case rdrCloseFile:
			return( doCloseFile( ) );


		/* Commit File - Subfunction 07h
		-------------------------------- */
		case rdrCommitFile:
			break;							// Pretend We Didn't See It


		/* Random Read - Subfunction 08h
		-------------------------------- */
		case rdrRead:
			return( doRead( ) );


		/* Random Write - Subfunction 09h
		--------------------------------- */
		case rdrWrite:
			return( doWrite( ) );


		/* Lock File Region - Subfunction 0Ah
		--------------------------------------- */
		case rdrLockRegion:
			break;							// Pretend We Didn't See It


		/* Unlock Region - Subfunction 0Bh
		---------------------------------- */
		case rdrUnlockRegion:
			break;							// Pretend We Didn't See It


		/* Get Disk Space - Subfunction 0Ch
		----------------------------------- */
		case rdrGetDiskSpace:
			return( doGetDiskSpace( ) );


		/* Set File Attribute - Subfunction 0Eh
		--------------------------------------- */
		case rdrSetFileAttrib:
			return( doSetFileAttrib( ) );


		/* Get File Attribute And Size - Subfunction 0Fh
		------------------------------------------------ */
		case rdrGetFileAttrib:
			return( doGetFileAttrib( ) );


		/* Rename File - Subfunction 11h
		-------------------------------- */
		case rdrRenameFile:
			return( doRenameFile( ) );


		/* Delete - Subfunction 13h
		--------------------------- */
		case rdrDeleteFile:
			return( doDeleteFile( ) );


		/* Open File - Subfunction 16h
		------------------------------ */
		case rdrOpenFile:
			return( doOpenFile( ) );


		/* Create File - Subfunction 17h
		-------------------------------- */
		case rdrCreateFile:
			return( doCreateFile( ) );


		/* Find First - Subfunction 1Bh
		------------------------------- */
		case rdrFindFirst:
			return( doFindFirst( ) );


		/* Find Next - Subfunction 1Ch
		------------------------------ */
		case rdrFindNext:
			return( doFindNext( ) );


		/* Close All Files For A Process - Subfunction 1Dh
		-------------------------------------------------- */
		case rdrCloseAll:
			return( 0 );					// Just Say Call Was Processed


		/* Seek From End Of File - Subfunction 21h
		------------------------------------------ */
		case rdrSeekFromEnd:
			return( doSeekFromEnd( ) );


		/* Process Died Notification - Subfunction 22h
		---------------------------------------------- */
		case rdrProcessDied:
			return( 0 );					// Just Say Call Was Processed


		/* Qualify Name
		--------------- */
		case rdrQualifyName:
			break;							// Pretend We Didn't See It


		/* Multi-Function Open - Subfunction 2Eh
		---------------------------------------- */
		case ifsSpecialOpen:
			return( doSpecialOpen( ) );

		default:
			break;
	}

	return( 1 );							// Pass The Buck
}

#ifdef __cplusplus
}
#endif

/****************************************************************************
 *                          Program MAIN Entry Point                        *
 ****************************************************************************/

/* Redirector Main Entry Point
------------------------------ */
int main( int argc, char **argv )
{
	int result;

	/* Say Hello!
	------------- */
	printf( "\nC-Phantom Redirector V%s\n", VERSION );
	printf( "Copyright (c) 1992-1995, Printing Communications Associates, Inc.\n" );
	printf( "All rights reserved.\n" );


	/* Check Argument Count
	----------------------- */
	if( argc != 2 )
	{
		printf( "Usage: CPHANTOM drive-letter\n" );
		exit( 1 );
	}


	/* Get And Verify Drive Letter And Drive Number
	----------------------------------------------- */
	drive_letter =  toupper( *argv[ 1 ] );
	drive_no =  drive_letter - '@';		// Drive Letter A: is Drive No. 1

	if( drive_no < 1 || drive_no > 27 )
	{
		/* Not In 'A' - 'Z' Range
		------------------------- */
		printf( "Usage: CPHANTOM drive-letter\n" );
		exit( 1 );
	}


	/* Get DOS Variables
	-------------------- */
	if( ( result = fsGetDOSVars() ) != 0 )
	{
		printf( "Error %d Installing Getting DOS Variables.\n", result );
		exit( 1 );
	}


	/* Initialize The Drive's CDS Structure
	--------------------------------------- */
	if( ( result = fsInitCDS( drive_no) ) != 0 )
	{
		printf( "Error %d Capturing Drive %c:\n", result, toupper( drive_letter ) );
		exit( 1 );
	}


	printf( "Phantom Drive Installed\n" );
	printf( "Drive    : %c:\n", toupper( drive_letter ) );


	/* Initialize Interrupt Intercept Assembler Glue
	------------------------------------------------ */
	ISRInit();


	/* Get A Free Software User Interrupt For Redirector Control
	------------------------------------------------------------ */
	_dos_setvect( 0x61, Int6XSvc );		// Just Pick 0x61 For Now

	printf( "Interrupt: %xh\n", 0x61 );

	/* Terminate And Stay Resident
	------------------------------ */
	_dos_keep( 0, (_SS + ( (_SP + 64 )/16) - _psp) );

	return( 0 );								// No Complaints Here...
}

